@@ -1,8 +1,10 @@
 package bundleuri
 
 import (
+	"context"
 	"fmt"
 	"path/filepath"
+	"strings"
 	"testing"
 
 	"github.com/prometheus/client_golang/prometheus/testutil"
@@ -13,12 +15,17 @@ import (
 	"gitlab.com/gitlab-org/gitaly/v18/internal/structerr"
 	"gitlab.com/gitlab-org/gitaly/v18/internal/testhelper"
 	"gitlab.com/gitlab-org/gitaly/v18/internal/testhelper/testcfg"
+	"google.golang.org/grpc/metadata"
 )
 
 func TestGenerationManager_Generate(t *testing.T) {
 	t.Parallel()
 
-	cfg := testcfg.Build(t)
+	const (
+		defaultNodeStorage = "gitaly-default"
+		defaultVirtualStorage
+	)
+	cfg := testcfg.Build(t, testcfg.WithStorages(defaultNodeStorage))
 	ctx := featureflag.ContextWithFeatureFlag(
 		testhelper.Context(t),
 		featureflag.BundleGeneration,
@@ -29,21 +36,39 @@ func TestGenerationManager_Generate(t *testing.T) {
 		desc            string
 		strategy        GenerationStrategy
 		setup           func(t *testing.T, repoPath string)
+		ctx             context.Context
 		expectedErr     error
+		expectedStorage string
 		expectFileExist bool
 	}{
 		{
-			desc:            "creates bundle successfully",
+			desc:            "creates bundle successfully with no virtual storage found in context",
 			expectFileExist: true,
+			ctx:             ctx,
 			setup: func(t *testing.T, repoPath string) {
 				gittest.WriteCommit(t, cfg, repoPath,
 					gittest.WithTreeEntries(gittest.TreeEntry{Mode: "100644", Path: "README", Content: "much"}),
 					gittest.WithBranch("main"))
 			},
+			expectedStorage: defaultNodeStorage,
+		},
+		{
+			desc:            "creates bundle successfully with virtual storage is found in context",
+			expectFileExist: true,
+			ctx: metadata.NewIncomingContext(ctx, metadata.New(map[string]string{
+				VirtualStorageKey: defaultVirtualStorage,
+			})),
+			setup: func(t *testing.T, repoPath string) {
+				gittest.WriteCommit(t, cfg, repoPath,
+					gittest.WithTreeEntries(gittest.TreeEntry{Mode: "100644", Path: "README", Content: "much"}),
+					gittest.WithBranch("main"))
+			},
+			expectedStorage: defaultVirtualStorage,
 		},
 		{
 			desc:            "fails with missing HEAD",
 			expectFileExist: false,
+			ctx:             ctx,
 			setup:           func(t *testing.T, repoPath string) {},
 			expectedErr:     structerr.NewFailedPrecondition("ref %q does not exist: %w", "refs/heads/main", fmt.Errorf("create bundle: %w", localrepo.ErrEmptyBundle)),
 		},
@@ -51,7 +76,7 @@ func TestGenerationManager_Generate(t *testing.T) {
 		t.Run(tc.desc, func(t *testing.T) {
 			t.Parallel()
 
-			repoProto, repoPath := gittest.CreateRepository(t, ctx, cfg, gittest.CreateRepositoryConfig{
+			repoProto, repoPath := gittest.CreateRepository(t, tc.ctx, cfg, gittest.CreateRepositoryConfig{
 				SkipCreationViaService: true,
 			})
 			repo := localrepo.NewTestRepo(t, cfg, repoProto)
@@ -67,18 +92,20 @@ func TestGenerationManager_Generate(t *testing.T) {
 
 			logger := testhelper.NewLogger(t)
 
-			manager, err := NewGenerationManager(ctx, sink, logger, nil, tc.strategy)
+			manager, err := NewGenerationManager(tc.ctx, sink, logger, nil, tc.strategy)
 			require.NoError(t, err)
 
-			err = manager.Generate(ctx, repo)
+			err = manager.Generate(tc.ctx, repo)
 			if tc.expectedErr != nil {
 				require.Equal(t, tc.expectedErr, err)
 				return
 			}
 
 			if tc.expectFileExist {
+				bundlePath := bundleRelativePath(tc.ctx, repo, "default")
+				require.True(t, strings.HasPrefix(bundlePath, tc.expectedStorage))
 				require.Equal(t, 1, testutil.CollectAndCount(manager, "gitaly_bundle_generation_seconds"))
-				require.FileExists(t, filepath.Join(sinkDir, bundleRelativePath(repo, "default")))
+				require.FileExists(t, filepath.Join(sinkDir, bundlePath))
 			}
 		})
 	}
@@ -147,9 +174,9 @@ func TestGenerationManager_GenerateWithStrategy(t *testing.T) {
 
 			if tc.expectFileExist {
 				require.Equal(t, 1, testutil.CollectAndCount(manager, "gitaly_bundle_generation_seconds"))
-				require.FileExists(t, filepath.Join(sinkDir, bundleRelativePath(repo, "default")))
+				require.FileExists(t, filepath.Join(sinkDir, bundleRelativePath(ctx, repo, defaultBundle)))
 			} else {
-				require.NoFileExists(t, filepath.Join(sinkDir, bundleRelativePath(repo, "default")))
+				require.NoFileExists(t, filepath.Join(sinkDir, bundleRelativePath(ctx, repo, defaultBundle)))
 			}
 		})
 	}
