@@ -11,14 +11,14 @@ import (
 	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/client_golang/prometheus/testutil"
 	"github.com/stretchr/testify/require"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/featureflag"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/git"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/git/gittest"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/git/quarantine"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/config"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/helper/text"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/testhelper"
-	"gitlab.com/gitlab-org/gitaly/v16/internal/testhelper/testcfg"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/featureflag"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/git"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/git/gittest"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/git/quarantine"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/gitaly/config"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/helper/text"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/testhelper"
+	"gitlab.com/gitlab-org/gitaly/v18/internal/testhelper/testcfg"
 )
 
 func TestObjectReader_reader(t *testing.T) {
@@ -194,6 +194,41 @@ func testObjectReaderObject(t *testing.T, ctx context.Context) {
 		require.EqualError(t, err, "current object has not been fully read")
 	})
 
+	t.Run("read submodule", func(t *testing.T) {
+		t.Parallel()
+		submoduleCommit := gittest.DefaultObjectHash.HashData([]byte("submodule commit"))
+		treeOID := gittest.WriteTree(t, cfg, repoPath, []gittest.TreeEntry{
+			{
+				Path:    ".gitmodules",
+				Mode:    "100644",
+				Content: fmt.Sprintf(`[submodule "sub"]\n\tpath = sub\n\turl = file://%s`, repoPath),
+			},
+			{
+				Path: "sub",
+				Mode: "160000",
+				OID:  submoduleCommit,
+			},
+		})
+
+		reader, err := newObjectReader(ctx, newRepoExecutor(t, cfg, repoProto), nil)
+		require.NoError(t, err)
+		defer reader.close()
+
+		commitID := gittest.WriteCommit(t, cfg, repoPath,
+			gittest.WithTree(treeOID),
+			gittest.WithMessage("Add submodule"),
+		)
+
+		submoduleRevision := fmt.Sprintf("%s:%s", commitID, "sub")
+
+		_, err = reader.Object(ctx, git.Revision(submoduleRevision))
+		if featureflag.GitMaster.IsEnabled(ctx) || !gittest.IsGitVersionLessThan(t, ctx, cfg, git.NewVersion(2, 51, 0, 0)) {
+			require.Equal(t, NotFoundError{Revision: submoduleCommit.String()}, err)
+		} else {
+			require.Equal(t, NotFoundError{Revision: commitID.String() + ":sub"}, err)
+		}
+	})
+
 	t.Run("read fails when partially consuming previous object", func(t *testing.T) {
 		reader, err := newObjectReader(ctx, newRepoExecutor(t, cfg, repoProto), nil)
 		require.NoError(t, err)
@@ -720,8 +755,9 @@ func TestObjectReader_logging(t *testing.T) {
 			SkipCreationViaService: true,
 		})
 
-		quarantineDir, err := quarantine.New(ctx, repoProto, logger, locator)
+		quarantineDir, cleanup, err := quarantine.New(ctx, repoProto, logger, locator)
 		require.NoError(t, err)
+		t.Cleanup(cleanup)
 
 		reader, err := newObjectReader(ctx, newRepoExecutor(t, cfg, quarantineDir.QuarantinedRepo()), nil)
 		require.NoError(t, err)
